/**
 * FocusAtMouse.java
 */
package UTEStatsViewer;
import java.awt.*;
import java.awt.event.*;
/**
 * Class that provides the function of having the keyboard focus be at
 * the mouse location.
 */
class FocusAtMouse extends MouseAdapter {
   private static FocusAtMouse instance = null;  // the singleton instance

   private Component focusTarget;


   /** Hide the default constructor. */
   private FocusAtMouse() {}


   /**
    * Create a MouseAdapter that sets the focus on a Component other
    * than the one that is the target of the mouseEntered event.  Used
    * when the focus is to be set to a Container that the target of the
    * mouseEvent is in.
    */
   FocusAtMouse(Component focusTarget) {
      this.focusTarget = focusTarget;
   }


   /**
    * Return the singleton of this class that sets the focus on the
    * Component that mouse enters. */
   public static FocusAtMouse getInstance() {
      if (instance == null) {
         instance = new FocusAtMouse();
      }
      return instance;
   }


   /** Request focus for the component when mouse enters it. */
   public void mouseEntered(MouseEvent e) {
      Component target = (this.focusTarget == null) ? e.getComponent() :
            this.focusTarget;
      target.requestFocus();
   }
}
