/*
 * NOTE:
 * CURRENTLY SLOG_API DOES NOT SUPPORT 'WHERE'
 * TURNING OFF -DNOWHERE CAUSES DISASTROUS RESULTS
 *
 * slogmerge: takes multiple UTE interval files, and creates an output file
 */
#include <stdio.h>
#include <math.h>
#include <sys/trchkid.h>
#include <values.h>
#include "mpitrcev.h"
#include "uteheader.h"
#include "event_define.h"
#include "slog.h"
#include "aix2ute.h"

typedef union {
  double dbl;
  int uint[2];
} dblint;

extern char taskbit[];
extern char threadTypeBit[];

/*
 * the following 3 variables are defined in utemerge.c
 */
dblint MINDIFF;
extern int max_num_markers;
extern table_format etable;
extern aixargs b_aix[];
extern aixargs e_aix[];

#define MILLION 1000000

static double end_time = 0.0;

double check_doubles(double s, double d) {
  double end, diff, min;
  int *ip, *jp;

  end = s+d;
  if (end + MINDIFF.dbl < end_time) {
    diff = end_time - end;
    fprintf(stderr,
      "Warning: end time %.18lf < last end time %.18lf := %.18lf\n",
      end, end_time, diff);
    min = MINDIFF.dbl;
    fprintf(stderr,"diff = %.18lf; min = %.18lf\n", diff, min);
    exit(4);
  } else if (end < end_time) {
    return(end_time - d);
  } else {
    end_time = end;
    return(s);
  }
}

/*
 * SLOG Argument array
 */
SLOG_arg_t irec_args[10];
SLOG_uint32 irec_Nargs;

slog_append_rec(SLOG_STREAM *slog, unsigned ev, unsigned be, long long *s, 
long long *d, unsigned nodeid, unsigned cpuid, unsigned tid, unsigned w, UTENODE *ptr) {
  SLOG_rectype_t rectype;
  SLOG_intvltype_t etype;
  SLOG_starttime_t stime;
  SLOG_duration_t duration;
  SLOG_bebit_t bbit;
  SLOG_bebit_t ebit;
  SLOG_nodeID_t node;
  SLOG_cpuID_t cpu;
  SLOG_threadID_t thread;
  SLOG_where_t where;
  SLOG_assoc_t assoc;
  SLOG_intvlrec_t *irec;
  long long ilong, lastEnd = 0;
  int i, j;

  rectype = SLOG_RECTYPE_STATIC_DIAG;
  etype = ev;

  ilong = *s + *d;
  if (ilong < lastEnd) {
    fprintf(stderr,"Error: end time %lld < last end time %lld\n",ilong,lastEnd);
    exit(3);
  } else
    lastEnd = ilong;
  ilong = *s;
  stime = (double) ilong /MILLION;
  ilong = *d;
  duration = (double) ilong / MILLION;
  stime = check_doubles(stime, duration);

  node = nodeid;
  cpu = cpuid;
  thread = tid;
  where = w;
  assoc = 0;

  if ((irec = SLOG_Irec_Create()) == (SLOG_intvlrec_t *)SLOG_FAIL) {
    fprintf(stderr,"Error: SLOG_Irec_Create() failed\n");
    exit(0);
  }
  if ( SLOG_Irec_SetMinRec( irec, rectype, etype, be >> 1, be & 0x01,
    stime, duration, node, cpu, thread, where ) == SLOG_FAIL) {
    fprintf(stderr,"Error: SLOG_Irec_SetMinRec() failed\n");
    exit(0);
  }

/*
 * Irec_SetArgs
 */
  irec_Nargs = 0;
  if (ev >= NUM_ENTRIES) {
    fprintf(stderr,"Error: event type %d unknown\n", ev);
    exit(2);
  }
  switch(ev) {
/*
 * the following records do not have instruction addresses
 */
  case _UTE_VERSION:
  case _TRCON:
  case _TB0_CLOCK:
  case _RUNNING:
  case _WAITALL_INFO:
  case _TESTALL_INFO_HDR:
  case _TESTALL_INFO:
  case _WAITSOME_INFO_HDR:
  case _WAITSOME_INFO:
  case _TESTSOME_INFO_HDR:
  case _TESTSOME_INFO:
  case _STARTALL_B_INFO:
  case _STARTALL_E_INFO:
  case _START_REQUEST:
  case _NEW_COMM_ID:
  case _NEW_COMM_RANKS:
  case _ENV_VARS:
  case _MPI_FAILURE:
  case _TRACE_TERMINATED:
  case _THREAD_IDS:
    switch(be) {
    case B_BEBIT:
      j = b_aix[ev].nargs;
      irec_Nargs = j;
      for (i = 0; i < j; i++) 
        irec_args[i] = *(int *)&(ptr->Eventbuf[MINBYTES+i*4]);
      break;
    case E_BEBIT:
      j = e_aix[ev].nargs ;
      irec_Nargs = j;
      for (i = 0; i < j; i++)
        irec_args[i] = *(int *)&(ptr->Eventbuf[MINBYTES+i*4]);
      break;
    case BE_BEBIT:
      j = b_aix[ev].nargs + e_aix[ev].nargs ;
      irec_Nargs = j;
      for (i = 0; i < j; i++)
        irec_args[i] = *(int *)&(ptr->Eventbuf[MINBYTES+i*4]);
      break;
    case CON_BEBIT:
      irec_Nargs = 0;
      break;
    }
    break;
  default:
    switch(be) {
    case B_BEBIT:
      j = b_aix[ev].nargs;
      irec_Nargs = j;
      for (i = 0; i < j; i++) 
        irec_args[i] = *(int *)&(ptr->Eventbuf[ADDBYTES+i*4]);
      break;
    case E_BEBIT:
      j = e_aix[ev].nargs ;
      irec_Nargs = j;
      for (i = 0; i < j; i++)
        irec_args[i] = *(int *)&(ptr->Eventbuf[ADDBYTES+i*4]);
      break;
    case BE_BEBIT:
      j = b_aix[ev].nargs + e_aix[ev].nargs ;
      irec_Nargs = j;
      for (i = 0; i < j; i++)
        irec_args[i] = *(int *)&(ptr->Eventbuf[ADDBYTES+i*4]);
      break;
    case CON_BEBIT:
      irec_Nargs = 0;
      break;
    }
    break;
  }
  if (irec_Nargs > 0)
    SLOG_Irec_SetArgs( irec, slog, irec_Nargs, irec_args );

  if ( SLOG_Irec_ToOutputStream( slog, irec ) == SLOG_FAIL ) {
    fprintf(stderr,"Error: SLOG_Irec_ToOutputStream() failed\n");
    exit(0);
  }
  SLOG_Irec_Free( irec );

}

#define MAX_ARGUMENTS 20
static char *name[MAX_ARGUMENTS];

add_profile(SLOG_STREAM *slogstr) {
  int i, j, k, rv;
  int loop, idx; 
  char *color;

  if (SLOG_PROF_Open(slogstr) == SLOG_FAIL) {
    fprintf(stderr, "Error: SLOG_PROF_Open() failed\n");
    exit(0);
  }
  for (loop = 0; loop < (_RUNNING+1)*4; loop++) {
    i = loop >> 2;
    j = loop & 0x3;
    if (j == 0) color = (char *)getRandomColorName();
    switch (i) {
    case _RUNNING:
/*
 * with no additional fields
 */
      if ((rv=SLOG_PROF_AddIntvlInfo(slogstr,i,j >> 1,j & 0x1,"Program_state",
        etable.record_names.name_array[etable.row[loop].name_index], 
        color, 0)) == SLOG_FAIL) {
      fprintf(stderr, "Error: SLOG_PROF_AddIntvlInfo(), type=%d be=%d\n",i,j);
        exit(0);
      }
      break;
/*
 * with no instruction address field
 */
    case _TB0_CLOCK:
    case _START_REQUEST:
    case _NEW_COMM_ID:
    case _NEW_COMM_RANKS:
    case _ENV_VARS:
    case _WAITALL_INFO:
    case _TESTALL_INFO:
    case _TESTALL_INFO_HDR:
    case _TESTSOME_INFO_HDR:
    case _TESTSOME_INFO:
    case _STARTALL_B_INFO:
    case _STARTALL_E_INFO:
    case _MPI_FAILURE:
    case _TRACE_TERMINATED:
    case _THREAD_IDS:
    case _UTE_VERSION:
      switch(j) {
      case B_BEBIT:
        k = b_aix[i].nargs;
        break;
      case E_BEBIT:
        k = e_aix[i].nargs ;
        break;
      case BE_BEBIT:
        k = b_aix[i].nargs + e_aix[i].nargs ;
        break;
      case CON_BEBIT:
        k = 0;
        break;
      }
/*
 * setup name[] pointers for arguments
 */
#ifdef DEBUG
      if (k > 0)
        fprintf(stderr,"%d %d", i, j);
#endif
      for (rv = 0; rv < k; rv++) {
        if (rv > (etable.row[loop].number_of_fields - 7)) {
          name[rv] = NULL;
          fprintf(stderr,"Warning: too many fields for rec=%d, be=%d, %d > %d specified\n", i, j, rv+7,
            etable.row[loop].number_of_fields);
        } else {
          name[rv] = etable.field_names.name_array[etable.row[loop].field[7+rv].name_index];
#ifdef DEBUG
          fprintf(stderr," %s", name[rv]);
#endif
        }
      }
#ifdef DEBUG
      if (k > 0)
        fprintf(stderr,"\n");
#endif

      if ((rv=SLOG_PROF_AddIntvlInfoWithArray(slogstr,i,j >> 1,j & 0x1,"Event_state",
        etable.record_names.name_array[etable.row[loop].name_index], 
        color, k, name)) == SLOG_FAIL) {
      fprintf(stderr, "Error: SLOG_PROF_AddIntvlInfo(), type=%d be=%d\n",i,j);
        exit(0);
      }
      break;
/*
 * these should not exist, since SLOG use separate rectype for them
 */
    case _USER_MARKER_INTERVAL:
    case _USER_MARKER_DEF:
/*
 * event records
 */
    case _PCONTROL:
      break;
    case _SEND:
    case _BSEND:
    case _RSEND:
    case _SSEND:
    case _ISEND:
    case _IBSEND:
    case _IRSEND:
    case _ISSEND:
    case _RECV:
    case _IRECV:
    case _SENDRECV:
    case _SENDRECV_REPLACE:
    default:
      switch(j) {
      case B_BEBIT:
        k = b_aix[i].nargs;
        idx = 1;
        break;
      case E_BEBIT:
        k = e_aix[i].nargs ;
        idx = 0;
        break;
      case BE_BEBIT:
        k = b_aix[i].nargs + e_aix[i].nargs ;
        idx = 1;
        break;
      case CON_BEBIT:
        k = 0;
        idx = 0;
        break;
      }
/*
 * k=number of fields
 * all field names must be passed using the SLOG_PROF_AddIntvlInfo()
 */
#ifdef DEBUG
      if (k > 0)
        fprintf(stderr,"%d %d", i, j);
#endif
      for (rv = 0; rv < k; rv++) {
        if (rv > (etable.row[loop].number_of_fields - 7 -idx)) {
          name[rv] = NULL;
          fprintf(stderr,"Warning: too many fields for rec=%d be=%d, %d > %d specified\n", i, j, rv+7+idx, 
            etable.row[loop].number_of_fields);
        } else {
          name[rv] = etable.field_names.name_array[etable.row[loop].field[rv+7+idx].name_index];
#ifdef DEBUG
          fprintf(stderr," %s", name[rv]);
#endif
        }
      }
#ifdef DEBUG
      if (k > 0)
        fprintf(stderr,"\n");
#endif

      if ((rv=SLOG_PROF_AddIntvlInfoWithArray(slogstr,i,j >> 1,j & 0x1,"MPI_state",
        etable.record_names.name_array[etable.row[loop].name_index], 
        color, k, name)) == SLOG_FAIL) {
      fprintf(stderr, "Error: SLOG_PROF_AddIntvlInfo(), type=%d be=%d\n",i,j);
        exit(0);
      }
      break;
    }
  }
  if ((i = SLOG_PROF_SetExtraNumOfIntvlInfos(slogstr, 
    4 * max_num_markers)) == SLOG_FAIL) {
    fprintf(stderr, "Error: SLOG_PROF_SetExtraNumIntvlInfos() failed\n");
    exit(0);
  }

/*
 * cannot call SLOG_PROF_Close()
 *
  if ((i = SLOG_PROF_Close(slogstr)) == SLOG_FAIL) {
    fprintf(stderr, "Error: SLOG_PROF_Close() failed\n");
    exit(0);
  }
 */
}

add_defs(SLOG_STREAM *slogstr) {
  int i, j, k;
  SLOG_intvltype_t rectype;
  int be;
  SLOG_N_assocs_t nassoc;
  SLOG_N_args_t nargs;

  if (SLOG_RDEF_Open(slogstr) == SLOG_FAIL) {
    fprintf(stderr, "Error: SLOG_RDEF_Open() failed\n");
    exit(0);
  }
  for (rectype = 0; rectype < _RUNNING+1; rectype++) {
    for (i = 0; i < 4; i++) {
      switch(rectype) {
/*
 * these two should not exist, since SLOG use separate rectype for them
 */
      case _USER_MARKER_INTERVAL:
      case _USER_MARKER_DEF:
        SLOG_RDEF_AddRecDef(slogstr, rectype, i >> 1, i & 0x1, 0, 0);
        break;
      case _SEND:
      case _BSEND:
      case _RSEND:
      case _SSEND:
      case _ISEND:
      case _IBSEND:
      case _IRSEND:
      case _ISSEND:
      case _RECV:
      case _IRECV:
      case _SENDRECV:
      case _SENDRECV_REPLACE:
      default:
        switch(i) {
        case B_BEBIT:
          j = b_aix[rectype].nargs;
          SLOG_RDEF_AddRecDef(slogstr, rectype, i >> 1, i & 0x1, 0, j);
          break;
        case E_BEBIT:
          j = e_aix[rectype].nargs ;
          SLOG_RDEF_AddRecDef(slogstr, rectype, i >> 1, i & 0x1, 0, j);
          break;
        case BE_BEBIT:
          j = b_aix[rectype].nargs + e_aix[rectype].nargs ;
          SLOG_RDEF_AddRecDef(slogstr, rectype, i >> 1, i & 0x1, 0, j);
          break;
        case CON_BEBIT:
          SLOG_RDEF_AddRecDef(slogstr, rectype, i >> 1, i & 0x1, 0, 0);
          break;
        }
        break;
      }
    }
  }
  if ((i = SLOG_RDEF_SetExtraNumOfRecDefs( 
    slogstr, 4 * max_num_markers)) == SLOG_FAIL) {
    fprintf(stderr, "Error: SLOG_RDEF_SetExtraNumOfRecDefs() failed\n");
    exit(0);
  }

/*
 * cannot call SLOG_RDEF_Close() as its document says
 *
  if ((i = SLOG_RDEF_Close(slogstr)) == SLOG_FAIL) {
    fprintf(stderr, "Error: SLOG_RDEF_Close() failed\n");
    exit(0);
  }
 */
}

int add_ttab(SLOG_STREAM *slogstr, UTENODE *uten, int nnodes) {
  int i, j, k;
  int ttype;
  SLOG_nodeID_t nodeid;
  SLOG_threadID_t tid;
  SLOG_OSprocessID_t ospid;
  SLOG_OSthreadID_t ostid;
  SLOG_appID_t appid;
  UTENODE *ptr;

  if (SLOG_TTAB_Open(slogstr) == SLOG_FAIL) {
    fprintf(stderr, "Error: SLOG_TTAB_Open() failed\n");
    exit(0);
  }
  k = 0;
  for (i = 0; i < nnodes; i++) {
    ptr = &uten[i];
    for (j = 0; j < MAX_THREADS_PER_NODE; j++) {
      if (ptr->head->logicthd[j].taskid != -1) {
        if (k < MAX_THREADS) {
          appid = ptr->head->logicthd[j].taskid;
          nodeid = ptr->head->logicthd[j].nodeid;
          tid = ptr->head->logicthd[j].ltid;
          ospid = ptr->head->logicthd[j].processid;
          ostid = ptr->head->logicthd[j].threadid;
          ttype = ptr->head->logicthd[j].threadType;
          if (ISBITSET(taskbit, appid) && ISBITSET(threadTypeBit, ttype)) {
            ptr->head->logicthd[j].selectMask = 0;
            if (SLOG_TTAB_AddThreadInfo(slogstr, 
              nodeid, tid, ospid, ostid, appid) == SLOG_FAIL) {
              fprintf(stderr, "Error: SLOG_TTAB_AddThreadInfo() failed\n");
              exit(0);
            }
          } else
            ptr->head->logicthd[j].selectMask = 1;
        }
        k++;
      }
    }
  }
  if ((i = SLOG_TTAB_Close(slogstr)) == SLOG_FAIL) {
    fprintf(stderr, "Error: SLOG_TTAB_Close() failed\n");
    exit(0);
  }
  return k;
}

#ifdef MAIN

#define MAX_NUM_MARKERS 32256
#define MILLION 1000000
#define MIN_RECORD_COUNT 100
#define MAX_RECORD_COUNT 50000000

FILE *infp;
char *infile = NULL;
char *rptfile = NULL;
char outfileslog[16] = "trcfile.slog";
char outfileute[16] = "trcfile.ute";
FILE *pfp, *outfp;
char *ProfileName = "profile.ute";
char *profile = NULL;

UTENODE *utenode;
char taskbit[MAX_NUM_TASKS/8];
#define MAX_THREAD_TYPES 8
char threadTypeBit[MAX_THREAD_TYPES/8] = { 0xff };

char *Progname;
int nnodes = 1;
int global = 1;
int utefile = 0;
int zeroDuration = 0;
table_format etable;
double thresh_time;   /* threshold value */
extern stackE *freestacke;

/*
 * variables for up to 20 SLOG user markers
 */
int max_num_markers = 20;
int reservedFactor = 2;

interval_header header;
frame_directory framedir;
/*
 * for SLOG, it's frame size in bytes and num of frames
 */
int num_of_records = 131072;
int num_of_frames = 128;
long long offset64, offset_org;

unsigned char *outmarker;

main(argc,argv)
int argc;
char *argv[];
{
  UTENODE *ptr;
  double dbl, adjustclk();
  SLOG_STREAM *slogstr;
  unsigned ui, index;
  unsigned nodeID, cpuID, threadID, beBits, eventType;
  int i, j, k, bitlength, bits;
  int where, ltid, cpu;
  char *name, *color;
  field_format *field;
  double start, dura, get_duration();
  long long ilong, ilongs, ilongd;
  thread_entry *tn;
  stackE *stacke;

  thresh_time = 0.00004;
  Progname = (char *) basename(argv[0]);
  getcmdline(argc,argv);
  outmarker = calloc(max_num_markers, sizeof(char));

  if (!global)
    fprintf(stderr,
      "Warning: merged trace does not have accurate timing with -g option\n");

/*
 * default UTEPROFILE
 */
  if (profile == NULL) {
    name = (char *)getenv("UTEPROFILE");
    if (name == NULL)
      profile = ProfileName;
    else
      profile = name;
  }
  if ((pfp = fopen(profile,"r")) == NULL) {
    perror("open profile");
    exit(0);
  }
  if ( read_etype_table(&etable, pfp, 0) < 0 )
    exit(-1);

/*
 * utenode is a (UTENODE *) pointer
 */
  if ((utenode = (UTENODE *) calloc(nnodes, sizeof(UTENODE))) == NULL) {
    fprintf(stderr,"Error: calloc() failed to allocate UTE NODES\n");
    exit(0);
  }

/*
 * utenode is a (UTENODE *) pointer
 */
  ptr = (UTENODE *) utetree_init(infile, nnodes, utenode, global, &etable);
  if (ptr == NULL) {
    fprintf(stderr,"Warning: no event to process\n");
    exit(0);
  }

/*
 * convert to SLOG
 */
    aixarg_init();
    MINDIFF.uint[0] = 0x3df00000;
    MINDIFF.uint[1] = 0x00000000;
    if ((slogstr = SLOG_OpenOutputStream(rptfile)) == NULL) {
      fprintf(stderr, "Error: open output file %s\n", rptfile);
      exit(0);
    }
    j = 0;
    for (i = 0; i < nnodes; i++)
      j += ptr->count;
    SLOG_SetFrameByteSize( slogstr, num_of_records);
    SLOG_SetIncreasingEndtimeOrder( slogstr );
    SLOG_SetPreviewName( slogstr, "/dev/null" );

/*
 * SLOG user markers
 */
    k = add_ttab(slogstr, &utenode[0], nnodes);
    num_of_frames = 1+ j*30 / (num_of_records - 32 - reservedFactor * k *
      SLOG_typesz[ min_IntvlRec ]);
    num_of_frames *= 2;
    SLOG_SetMaxNumOfFramesPerDir( slogstr, num_of_frames);
    SLOG_SetFrameReservedSpace( slogstr, 
      reservedFactor * k * SLOG_typesz[ min_IntvlRec ]);
    add_profile(slogstr);
    add_defs(slogstr);
  
    do {
      index = *(unsigned *)ptr->Eventbuf >> (32 - etable.log_table_size);
      bitlength = 0;
      ui = *(unsigned *)&(ptr->Eventbuf);
      k = etable.row[index].number_of_sfields;
      for (i = 0; i < 7; i++) {
        field = (field_format *)get_field_format(&etable, index, i);
        if (field == NULL) {
          fprintf(stderr, "Error: NULL field format while parsing for field \
%d of index %d at node %d\n", i, index, ptr->nodeid);
          exit(0);
        }
        bits = (field->log_max_length == 0) ? 64 : field->log_max_length;
        name = (char *) get_field_name(&etable, index, i);
/*
 * hard-coded stuff
 * since SLOG does not know UTE records and requires individual fields
 */
        switch(i) {
        case 0:
          eventType = ui >> (32 - bits);
          break;
        case 1:
          beBits = (ui << bitlength) >> (32 - bits);
          break;
        case 2:
          nodeID = (ui << bitlength) >> (32 - bits);
          break;
        case 3:
          cpuID = (ui << bitlength) >> (32 - bits);
          break;
        case 4:
          threadID = *(ptr->Eventbuf + 4);
          break;
        case 5:
/* 
 * get_next_UTE() stores the endtime in (Asecond, Ananosecond)
 */
          start = floor((double)0.001 * ptr->Ananosecond + 
            ptr->Asecond * MILLION);
          break;
        case 6:
          dura = floor(get_duration(ptr->Eventbuf) * MILLION);
          ilongd = (long long) dura;
          start -= dura;
          ilongs = (long long) start;

          if (k > 7) {
            name = (char *)get_field_name(&etable, index, 7);
            if (!strcmp(name, "where"))
              where = *(int *)&(ptr->Eventbuf[MINBYTES]);
            else
              where = 0;
          } else
            where = 0;
          switch(eventType) {
          case _USER_MARKER_DEF:
/*
 * marker code may not be in ascending order, even for sequentially converted
 * UTE files
 */
            j = *(int *)&(ptr->Eventbuf[ADDBYTES]);
            if (j < max_num_markers && outmarker[j] == 0) {

              for (k = 0; k < 4; k++) {
/*
 * currently redundant SLOG API
 */
                if (k == 0) color = (char *)getRandomColorName();
                if (SLOG_PROF_AddExtraIntvlInfo(slogstr, _MARKER_BASE+j,
                  k >> 1, k & 0x1, "User_Marker", &(ptr->Eventbuf[ADDBYTES+5]), 
                  color) == SLOG_FAIL) {
                  fprintf(stderr, 
                    "Error: SLOG_PROF_AddExtraIntvlInfo() failed for type=%d\n",
                    i);
                  exit(0);
                }

                if (SLOG_RDEF_AddExtraRecDef(slogstr, _MARKER_BASE+j, k >> 1, 
                  k & 0x1, 0, 0) == SLOG_FAIL) {
                  fprintf(stderr, 
                    "Error: SLOG_RDEF_AddExtraRecDef() failed for type=%d\n",
                    _MARKER_BASE+j);
                  exit(0);
                }
              }
              outmarker[j] = 1;
            }
            break;
          case _USER_MARKER_INTERVAL:
            if (beBits == 3)
              k = 2;
            else if (beBits > 0)
              k = 1;
            else
              k = 0;
            j = *(int *)&(ptr->Eventbuf[MINBYTES+4*k]);
            if ( !(ptr->head->logicthd[get_ltid(ptr->Eventbuf)].selectMask) )
              slog_append_rec(slogstr,j+_MARKER_BASE, beBits, &ilongs, 
                &ilongd, nodeID, cpuID, threadID, where, ptr);
            break;
          case _PCONTROL:
          case _START_REQUEST:
          case _ENV_VARS:
          case _NEW_COMM_ID:
          case _NEW_COMM_RANKS:
          case _WAITALL_INFO:
          case _TESTALL_INFO_HDR:
          case _TESTALL_INFO:
          case _WAITSOME_INFO_HDR:
          case _WAITSOME_INFO:
          case _TESTSOME_INFO_HDR:
          case _TESTSOME_INFO:
          case _STARTALL_B_INFO:
          case _STARTALL_E_INFO:
          case _TRACE_TERMINATED:
          case _UTE_VERSION:
          case _TRCON:
            break;
          case _TB0_CLOCK:
          case _RUNNING:
          default:
            if ( !(ptr->head->logicthd[get_ltid(ptr->Eventbuf)].selectMask) )
              slog_append_rec(slogstr, eventType, beBits, &ilongs, 
                &ilongd, nodeID, cpuID, threadID, where, ptr);
            break;
          }

/* 
 * if (nnodes > 1 && global && get_etype(ptr->Eventbuf, &etable) == _TB0_CLOCK)
          dbl = adjustclk(ptr,thresh_time); 
 */
          break;
        }
        bitlength += bits;
      }

    } while ((ptr = (UTENODE *)get_next_UTE()) != NULL);

    if (SLOG_CloseOutputStream(slogstr) == SLOG_FAIL) {
      fprintf(stderr,"Error: SLOG_CLoseOutputStream() failed\n");
      exit(0);
    }
 
  exit(0);
}

static getcmdline(argc,argv)
char *argv[];
{
  int c;
  int n;
  int len;
  int i, range;
  char *cp;
  char *options;

  range = 0;
  options = "#?zgum:r:c:f:p:n:o:t:s::";
  while((c = getarg(argc,argv,options)) != EOF) {
    switch(c) {
    case 'z':
      zeroDuration = 1;
      break;
    case 'n':
      if ((nnodes = atoi(optarg)) < 1) {
        fprintf(stderr,"Error: number of nodes %d less than 1\n", nnodes);
        exit(0);
      } else if (nnodes > MAX_NUM_NODES) {
        fprintf(stderr,"Error: number of nodes %d larger than %d\n", 
          nnodes, MAX_NUM_NODES);
        exit(0);
      }
      break;
    case 'r':
      i = atoi(optarg);
      if (i < 1) {
        fprintf(stderr,"Error: reserved factor for reserved pseudo records in SLOG should be at least 1\n", i);
        exit(0);
      }
      reservedFactor = i;
      break;
    case 'm':
      i = atoi(optarg);
      if (i > MAX_NUM_MARKERS) {
        fprintf(stderr,"Warning: max number of user markers %d larger than %d\nmax number of user markers set to %d\n", i, MAX_NUM_MARKERS, MAX_NUM_MARKERS);
        i = MAX_NUM_MARKERS;
      } else if (i < 1) {
        fprintf(stderr,"Error: non-positive max number of user markers%d\n"
, i);
        exit(0);
      }
      max_num_markers = i;
      break;
    case 'f':
      i = atoi(optarg);
      if (i > MAX_NUM_OF_FRAMES) {
        fprintf(stderr,"Option -f ignored: number of frames per directory %d larger than %d\nnumber of frames per directory set to %d\n",
          i, MAX_NUM_OF_FRAMES, MAX_NUM_OF_FRAMES);
        i = MAX_NUM_OF_FRAMES;
      } else if (i < 1) {
        fprintf(stderr,"Error: non-positive number of frames per directory %d\n"
, i);
        exit(0);
      }
      num_of_frames = i;
      break;
    case 'c':
      i = atoi(optarg);
      if (i < MIN_RECORD_COUNT)
        fprintf(stderr,"Option -c ignored: record count per frame %d too small\n", i);
      else if (i > MAX_RECORD_COUNT)
        fprintf(stderr,"Option -c ignored: record count per frame %d too large\n"
, i);
      else
        num_of_records = i;
      break;
    case 'u':
      utefile = 1;
      break;
    case 'g':
      global = 0;
      break;
    case 's':
      range = 1;
      memset(taskbit, 0, MAX_NUM_TASKS/8);
      setrange(optarg);
      break;
    case 't':
      memset(threadTypeBit, 0, MAX_THREAD_TYPES/8);
      settypes(optarg);
      break;
    case 'o':
      rptfile = optarg;
      break;
    case 'p':
      profile = optarg;
      break;
    case '?':
    case '#':
    default:
      usage();
      break;
    }
  }
  if (optind == argc-1) {
    infile = argv[optind];
  } else
    usage();
  if (rptfile == NULL) {
    if (utefile)
      rptfile = outfileute;
    else
      rptfile = outfileslog;
  }
  if (!range)
    memset(taskbit, 0xff, MAX_NUM_TASKS/8);
}

setrange(char *hookbuf) {
  char *cp, *c;
  int i, j, k;
  char *delims;

  if(hookbuf == NULL)
    return;
  delims = ", ";
  for(cp = (char *)strtok(hookbuf,delims); cp; cp = (char *)strtok(0,delims)) {
    if ((c = (char *)strchr(cp, '-')) == NULL) {
      i = atoi(cp);
      SETBIT(taskbit, i);
    } else {
      j = atoi(cp);
      i = atoi(c+1);
      for (k = j; k <= i; k++) {
        SETBIT(taskbit, k);
      }
    } 
  }
}

settypes(char *hookbuf) {
  char *cp, *c;
  int i, j, k;
  char *delims;

  if(hookbuf == NULL)
    return;
  delims = ", ";
  for(cp = (char *)strtok(hookbuf,delims); cp; cp = (char *)strtok(0,delims)) {
    if ((c = (char *)strchr(cp, '-')) == NULL) {
      i = atoi(cp);
      if (i < MAX_THREAD_TYPES)
        SETBIT(threadTypeBit, i);
    } else {
      j = atoi(cp);
      i = atoi(c+1);
      for (k = j; k <= i; k++) {
        if (k < MAX_THREAD_TYPES)
          SETBIT(threadTypeBit, k);
      }
    } 
  }
}

usage() {
printf("usage: %s [-?#] [-c #_records] [-r SLOG_factor] [-m #_markers] [-s range] [-n #_nodes] [-p profile] [-o out_file] in_prefix\n",Progname);
printf("\
-? or -#     Usage\n\
-c number    Specify the number of bytes in each SLOG frame\n\
-r factor    Specify the reserved factor for pseudo records (default=2)\n\
-m #_markers Specify the maximum number of user markers (default=20)\n\
-n #_nodes   Specify the number of nodes (default=1)\n\
-p profile   Specify the binary profile (default=profile.ute)\n\
-o out_file  Specify output file name (default=trcfile.slog or trcfile.ute)\n\
-s range     Specify the tasks to be merged\n\
             task IDs could be separated by comma (,) or hyphen (-)\n\ 
             example: -s 0,2,4,5-7\n\
-t threadType Specify thread types to be merged. default=all thread types\n\
             thread types are separated by comma (,)\n\ 
             example: '-t 0' to choose only MPI threads\n\
in_prefix    Specify input file name prefix\n", Progname);
   exit(0);
/*
 * future options
-t threshold Specify adjustment threshold in micro-seconds (defaults to 40)\n\
 */
}

/*
 * some other common routines
 */
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define ARGSIZE 10240

getarg(argc,argv,options)
char *argv[];
char *options;
{
	int c;
	int len;
	char *cp,*tokp;
	char argbuf[ARGSIZE+4];

	c = getopt(argc,argv,options);                        /* call getopt */
	if(!iscmplx(c,options)) {
		return(c);  /* not a '::', just return what getopt() returns */
	}

	cp = optarg;
	tokp = (char *) strtok(optarg,", \t");      /* point to first text token */

	/* the following loop will remove any ", \t" characters and separate */
	/* each text token by a space character                              */
	do {
		len = strlen(tokp);
		cp = (char *)strcpy(cp,tokp);
		cp += len;
		*cp++ = ' ';
	} while((tokp = (char *)strtok(0,", \t")) != NULL);

	len = MIN(cp-optarg,ARGSIZE);                    /* length of string */
	memcpy(argbuf,optarg,len);                       /* append to buffer */

	argbuf[len-1] = '\0';                                 /* 0-terminate */
	if((optarg = (char *)malloc(len+1)) == NULL) {        /* allocate memory */
		perror("getarg");                         /* for contiguous */
		exit(1);                                  /* single argument */
	}
	strcpy(optarg,argbuf);                           /* copy into buffer */
	return(c);                                          /* return option */
}

/*
 * return non-zero if the 'c0' option is a :: option.
 */
iscmplx(c0,options)
char *options;
{
	char *cp;
	char c;

	cp = options;
	while(c = *cp++) {
		if(c == c0) {
			if(*cp++ == ':' && *cp++ == ':')
				return(1);
			return(0);
		}
	}
	return(0);
}

/*
 * Print system call error message and terminate 
 */
void syserr(msg)
char *msg;
{
	extern int errno, sys_nerr;
	extern char *sys_errlist[];

	fprintf(stderr, "ERROR: %s (%d", msg, errno);
	if (errno > 0 && errno < sys_nerr)
		fprintf(stderr, "; %s)\n", sys_errlist[errno]);
	else
		fprintf(stderr, ")\n");
	exit(0);
}

#endif
